#ifndef _ARITHMETIC_ASM_
#define _ARITHMETIC_ASM_

;=======================================================================;
; Arithmetic functions                                                  ;
;                                                                       ;
; Target mcu:   Atmel AVR ATmega8                                       ;
; Mcu clock:    12 Mhz                                                  ;
; Author:       Viliam Sabol (sabolv@gmail.com)                         ;
;                                                                       ;
;=======================================================================;

.include "m8def.inc"

	; 24bit * 8bit unsigned multiplication
	; multiplicand is stored in registers a0:2
	; multiplier is stored in register b
	; 32bit result of multiplication is pushed to registers res0:3
	; multiplicand register b is not preserved
	; routine uses registers r16 : r22
	; register for internal counter r23	is restored on return
	.def a0		= r16
	.def a1 	= r17
	.def a2 	= r18
	.def b	 	= r19
	.def res0	= r19
	.def res1 	= r20
	.def res2	= r21
	.def res3 	= r22
	.def cnt 	= r23

	MUL24U8:
		push cnt	
		clr	res3			; clear result bytes
		clr	res2
		clr res1
		ldi	cnt, 8			; init bit counter
		lsr	b				; shift first bit of multiplier to carry

		_MUL24U8_1:
		brcc _MUL24U8_2		; if bit from multiplier shifted to carry is set
		add	res1, a0		; add multiplicand to high bytes of result
		adc	res2, a1	
		adc	res3, a2

		_MUL24U8_2:	
		ror	res3			; result with multiplier in last byte
		ror	res2
		ror	res1
		ror	res0			; next bit of multiplier is shifted to carry
		dec	cnt				; decrement bit counter
		brne _MUL24U8_1		; loop if not all bits of multiplier are shifted
		pop cnt
		ret

	; 24bit / 24 bit unsigned division
	; dividend is stored in registers a0:2
	; divisor is stored in registers b0:2
	; result of division is pushed to dividend registers
	; a0:2 / b0:2 = a0:2, reminder in rem0:2
	; routine uses registers r13 : r22
	; register for internal counter r22 is restored on return
	; b0:2 registers are not changed

	.def a0		= r16
	.def a1 	= r17
	.def a2 	= r18
	.def b0 	= r19
	.def b1 	= r20
	.def b2 	= r21
	.def rem0 	= r13
	.def rem1 	= r14
	.def rem2 	= r15
	.def cnt 	= r22

	DIV24U:
		push cnt
		clr	rem0		; clear remainder
		clr rem1
		sub	rem2, rem2	; clear remainder high byte and carry
		ldi	cnt , 25	; init bit counter
		
		_DIV24U_1:
		rol	a0			; shift left dividend, last bit is in carry
		rol a1
		rol	a2
		dec	cnt			; decrement bit counter
		brne _DIV24U_2	; move on next bit if not finished
		pop cnt
		ret				; otherwise return

		_DIV24U_2:	
		rol	rem0		; shift dividend from carry into remainder
		rol rem1
		rol	rem2	
		sub	rem0, b0	; try to subtract divisor from remainder
		sbc	rem1, b1
		sbc rem2, b2
		brcc _DIV24U_3	; if result was not negative write 1 to result
						; if result is negative undo subtraction from 
						; remainder and write 0 to result
		add	rem0, b0	; add divisor back to remainder
		adc rem1, b1
		adc rem2, b2	
		clc				; clear carry bit that will be shifted to result
		rjmp _DIV24U_1	; else

		_DIV24U_3:
		sec				; set carry bit that will be shifted to result
		rjmp _DIV24U_1

#endif  /* _ARITHMETIC_ASM_ */
