#ifndef _AUXILIARY_ASM_
#define _AUXILIARY_ASM_

;=======================================================================;
; Auxiliary functions                                                   ;
;                                                                       ;
; Target mcu:   Atmel AVR ATmega8                                       ;
; Mcu clock:    12 Mhz                                                  ;
; Author:       Viliam Sabol (sabolv@gmail.com)                         ;
;                                                                       ;
;=======================================================================;

.include "m8def.inc"

#define _BV(X) (1 << X)	; bit value macro

;=======================================================================;
; EEPROM                                                                ;
;=======================================================================;

	; Writes data stored in r16 register to EEPROM address in r17:r18.
	; Interrupts have to be disabled during eeprom write operation!
	EEPROM_WRITE:		
		sbic EECR,EEWE	; Wait for completion of previous write
		rjmp EEPROM_WRITE
		out EEARH, r18	; Set up address (r18:r17) in address register
		out EEARL, r17
		out EEDR,r16	; Write data (r16) to data register
		sbi EECR,EEMWE	; Write logical one to EEMWE
		sbi EECR,EEWE	; Start eeprom write by setting EEWE
		ret

	; Read data from EEPROM address in r17:r18 to r16 register.
	; Interrupts have to be disabled during eeprom read operation!
	EEPROM_READ:
		sbic EECR,EEWE	; Wait for completion of previous write
		rjmp EEPROM_READ
		out EEARH, r18	; Set up address (r18:r17) in address register
		out EEARL, r17
		sbi EECR,EERE	; Start eeprom read by writing EERE
		in r16,EEDR	; Read data from data register
		ret

	; Load value from eeprom address to r16 register content
	; Interrupts have to be disabled during read operation.
	.macro eepromread ; address
		ldi r17, low(@0)
		ldi r18, high(@0)
		rcall EEPROM_READ
	.endmacro

	; Store value to eeprom address from r16 register content
	; Interrupts have to be disabled during write operation.
	.macro eepromwrite ; address
		ldi r17, low(@0)
		ldi r18, high(@0)
		rcall EEPROM_WRITE
	.endmacro

;=======================================================================;
; SRAM                                                                  ;
;=======================================================================;

	; Load byte to register from variable(defined by
	; .byte directive) in sram, using register Z
	.macro ldsram ; rX, var
		ldi ZL, LOW(@1)
        ldi ZH, HIGH(@1)
		ld @0, Z
	.endmacro

	; Store byte from register to variable(defined by
	; .byte directive) in sram, using register Z
	.macro stsram ; var, rX
		ldi ZL, LOW(@0)
        ldi ZH, HIGH(@0)
		st Z, @1
	.endmacro

	; Store byte value to variable(defined by
	; .byte directive) in sram, using register Z
	.macro stsrami ; var, X
		ldi ZL, LOW(@0)
        ldi ZH, HIGH(@0)
		push r16
		ldi r16, @1
		st Z, r16
		pop r16
	.endmacro

	; Store 24bit integer value in registers regLo:regHi
	; to specified address(defined by .byte 3 directive) in sram.
	.macro stsram24u ; address regLo, regMi, regHi
		ldi ZL, LOW(@0)
        ldi ZH, HIGH(@0)
		st Z+, @1
		st Z+, @2
		st Z,  @3
	.endmacro

	; Load 24bit integer value from sram at specified address
	; (defined by .byte 3 directive), to registers regLo:regHi
	.macro ldsram24u ; regLo, regMi, regHi, address
		ldi ZL, LOW(@3)
        ldi ZH, HIGH(@3)
		ld @0, Z+
		ld @1, Z+
		ld @2, Z
	.endmacro

	; Load 24bit integer value from sram at specified address
	; (defined by .byte 3 directive) + register with offset value
	; to registers regLo:regHi
	.macro ldsram24uor ; regLo, regMi, regHi, address, regOffset
		ldi ZL, LOW(@3)
        ldi ZH, HIGH(@3)
		push r16
		clr r16
		add ZL, @4		
		adc ZH, r16
		pop r16
		ld @0, Z+
		ld @1, Z+
		ld @2, Z
	.endmacro

	; Load 24bit integer into registers
	.macro ldi24u ; regLo, regMi, regHi, value
		ldi @0,   low(@3)
		ldi @1,  high(@3)
		ldi @2, byte3(@3)
	.endmacro

	; Load 16bit integer into registers
	.macro ldi16u ; regLo, regHi, value
		ldi @0,   low(@2)
		ldi @1,  high(@2)
	.endmacro

;=======================================================================;
; Strings                                                               ;
;=======================================================================;

	; Conversion routine from unsigned 24bit integer to 8 ascii characters
	; that represent unsigned number on input
	; input integer lo:hi
	; output chars r7:r0
	; this routine changes input registers
	.def lo 	= r16
	.def mi 	= r17
	.def hi 	= r18
	
	; Helper macro generates one digit from input number. Counts
	; number division of number in loreg:hireg by divcnt
	#define _INT2STR24_DIGIT(label, divcnt, outreg, loreg, mireg, hireg) \
		label:						\
		subi loreg, low(divcnt)		\
		sbci mireg, high(divcnt)	\
		sbci hireg, byte3(divcnt)	\
		inc outreg					\
		brcc label					\
		dec outreg					\
		subi loreg, low(-divcnt)	\
		sbci mireg, high(-divcnt)	\
		sbci hireg, byte3(-divcnt)	\

	INT2STR24:
		push lo
		push mi
		push hi
		push r19

		ldi r19, '0'	; fill zeros to output registers
		mov r0, r19
		mov r1, r19
		mov r2, r19
		mov r3, r19
		mov r4, r19
		mov r5, r19
		mov r6, r19
		mov r7, r19
		
		cpi hi, 0
		brne _INT2STR24_CONV
		cpi mi, 0
		brne _INT2STR24_5
		cpi lo, 0
		brne _INT2STR24_3
		rjmp _INT2STR24_RET

		_INT2STR24_CONV:
		_INT2STR24_DIGIT(_INT2STR24_8,10000000, r7, lo, mi, hi)
		_INT2STR24_DIGIT(_INT2STR24_7, 1000000, r6, lo, mi, hi)
		_INT2STR24_DIGIT(_INT2STR24_6,  100000, r5, lo, mi, hi)
		_INT2STR24_DIGIT(_INT2STR24_5,   10000, r4, lo, mi, hi)
		_INT2STR24_DIGIT(_INT2STR24_4,    1000, r3, lo, mi, hi)
		_INT2STR24_DIGIT(_INT2STR24_3,     100, r2, lo, mi, hi)
		_INT2STR24_DIGIT(_INT2STR24_2,      10, r1, lo, mi, hi)
		
		add r0, lo

		_INT2STR24_RET:
		pop r19
		pop hi
		pop mi
		pop lo
		ret

	; Conversion routine from unsigned 16bit integer to 5 ascii characters
	; that represent unsigned number on input
	; input integer lo:hi
	; output chars r4:r0
	; this routine changes input registers
	.def lo 	= r16
	.def hi 	= r17
	
	; Helper macro generates one digit from input number. Counts
	; number division of number in loreg:hireg by divcnt
	#define _INT2STR16_DIGIT(label, divcnt, outreg, loreg, hireg) \
		label:						\
		subi loreg, low(divcnt)		\
		sbci hireg, high(divcnt)	\
		inc outreg					\
		brcc label					\
		dec outreg					\
		subi loreg, low(-divcnt)	\
		sbci hireg, high(-divcnt)	\

	INT2STR16:
		push lo
		push hi
		push r18

		ldi r18, '0'	; fill zeros to output registers
		mov r0, r18
		mov r1, r18
		mov r2, r18
		mov r3, r18
		mov r4, r18
		
		cpi hi, 0
		breq _INT2STR16_3

		_INT2STR16_DIGIT(_INT2STR16_5, 10000, r4, lo, hi)
		_INT2STR16_DIGIT(_INT2STR16_4,  1000, r3, lo, hi)
		_INT2STR16_DIGIT(_INT2STR16_3,   100, r2, lo, hi)
		_INT2STR16_DIGIT(_INT2STR16_2,    10, r1, lo, hi)
		
		add r0, lo

		pop r18
		pop hi
		pop lo
		ret

	; Removes leading zeros from 24 bit integer in ascii code stored in
	; registers r7:r0. Leading zeros are replaced with spaces
	; this routine changes input registers
	; lzero register contains maximum number of digits to be replaced	
	.def zchr = r17
	.def lzero = r16

	; Helper macro that replaces zero to space if lzero flag is not set
	; and proceeds on the next digit label
	#define _REMLZEROS_DIGIT(lblname, lblnum, nextlblnum, outreg) \
		lblname ## lblnum:				\
		cp outreg, zchr					\
		brne lblname ## 0				\
		tst lzero						\
		breq lblname ## 0				\
		sub outreg, r18					\
		dec lzero						\
		rjmp lblname ## nextlblnum		\

	REMLZEROS24:
		push r16
		push r17
		push r18
		
		ldi lzero, 8
		ldi zchr, '0'
		ldi r18, '0' - ' '
		
		_REMLZEROS_24:
		_REMLZEROS_DIGIT(_REMLZEROS_, 7, 6, r7)
		_REMLZEROS_DIGIT(_REMLZEROS_, 6, 5, r6)
		_REMLZEROS_DIGIT(_REMLZEROS_, 5, 4, r5)
		_REMLZEROS_16:
		_REMLZEROS_DIGIT(_REMLZEROS_, 4, 3, r4)
		_REMLZEROS_DIGIT(_REMLZEROS_, 3, 2, r3)
		_REMLZEROS_DIGIT(_REMLZEROS_, 2, 1, r2)
		_REMLZEROS_DIGIT(_REMLZEROS_, 1, 0, r1)

		_REMLZEROS_0:
		pop r18
		pop r17
		pop r16
		ret

	; Removes leading zeros from 16 bit integer in ascii code stored in
	; registers r4:r0. Leading zeros are replaced with spaces
	; this routine changes input registers	
	; lzero register contains maximum number of digits to be replaced	
		
	REMLZEROS16:
		push r16
		push r17
		push r18
		
		ldi lzero, 5
		ldi zchr, '0'
		ldi r18, '0' - ' '
		rjmp _REMLZEROS_16

	; Removes leading zeros from 24 bit integer with last two decimal digits 
	; in ascii code stored in registers r7:r0. Leading zeros are replaced 
	; with spaces this routine changes input registers	
	; lzero register contains maximum number of digits to be replaced	
		
	REMLZEROS24_FP2:
		push r16
		push r17
		push r18
		
		ldi lzero, 5
		ldi zchr, '0'
		ldi r18, '0' - ' '
		rjmp _REMLZEROS_24

	; Removes leading zeros from 24 bit integer with last digit as decimal
	; in ascii code stored in registers r7:r0. Leading zeros are replaced 
	; with spaces this routine changes input registers	
	; lzero register contains maximum number of digits to be replaced	
		
	REMLZEROS24_FP1:
		push r16
		push r17
		push r18
		
		ldi lzero, 6
		ldi zchr, '0'
		ldi r18, '0' - ' '
		rjmp _REMLZEROS_24


#endif /* _AUXILIARY_ASM_ */
