#ifndef _LCD_ASM_
#define _LCD_ASM_

;=======================================================================;
; Library for LCD modules with HD44780 comatible driver                 ;
;                                                                       ;
; Target mcu:   Atmel AVR ATmega8                                       ;
; Mcu clock:    12 Mhz                                                  ;
; Author:       Viliam Sabol (sabolv@gmail.com)                         ;
;                                                                       ;
;=======================================================================;

.include "m8def.inc"

	; LCD connection definitions
	.equ LCD_RS_PIN 	= PB1
	.equ LCD_RS_PORT 	= PORTB
	.equ LCD_RS_DDR		= DDRB
	.equ LCD_RS_MSK 	= _BV(LCD_RS_PIN)

	.equ LCD_EN_PIN 	= PB2
	.equ LCD_EN_PORT 	= PORTB
	.equ LCD_EN_DDR 	= DDRB
	.equ LCD_EN_MSK 	= _BV(LCD_EN_PIN)

	.equ LCD_DATA_PORT	= PORTC
	.equ LCD_DATA_DDR 	= DDRC
	.equ LCD_DATA_MSK 	= _BV(PC0)|_BV(PC1)|_BV(PC2)|_BV(PC3)

	; LCD DDRAM Address
	.equ LCD_DDRAM_LINE0 = 0
	.equ LCD_DDRAM_LINE1 = 64

	; LCD delay definitions
	.equ LCD_DELAY_MS_INIT1		= 20
	.equ LCD_DELAY_MS_INIT2 	= 10
	.equ LCD_DELAY_US_INIT3		= 100
	.equ LCD_DELAY_US_INSTR 	= 40
	.equ LCD_DELAY_MS_CLR_RET 	= 4

	; LCD instruction bytes:

	; Init instruction
	.equ LCD_INSTR_INIT_FNSET8B 	= 0b00000011
	.equ LCD_INSTR_INIT_FNSET4B 	= 0b00000010

	; Function set: 0 0 1 DL N F * *
	.equ LCD_INSTR_FNSET_4B_2L_58F 	= 0b00101000

	; Display on/off control: 0 0 0 0 1 D C B
	.equ LCD_INSTR_DCTRL_OFF_NC_NB 	= 0b00001000
	.equ LCD_INSTR_DCTRL_ON_NC_NB 	= 0b00001100

	; Clear display: 0 0 0 0 0 0 0 1
	.equ LCD_INSTR_CLRDISP 			= 0b00000001

	; Entry mode set: 0 0 0 0 0 1 I/D S
	.equ LCD_INSTR_ENTRY_INC_NS		= 0b00000110
	
	; DDRAM addres set mask
	.equ LCD_INSTR_SET_DDRAM		= 0b10000000

	; CGRAM addres set mask
	.equ LCD_INSTR_SET_CGRAM		= 0b01000000

;=======================================================================;
; LCD functions                                                         ;
;=======================================================================;

	; Strobes LCD enable bit to commit data on LCD data links
	LCD_STROBE:		
		sbi LCD_EN_PORT, LCD_EN_PIN
		nop
		cbi LCD_EN_PORT, LCD_EN_PIN		
		ret

	; Initialize LCD connection ports
	LCD_PORT_INIT:
		push r16
		
		; init RS pin
		in r16, LCD_RS_DDR
		ori r16, LCD_RS_MSK
		out LCD_RS_DDR, r16

		; init DATA pins
		in r16, LCD_DATA_DDR
		ori r16, LCD_DATA_MSK
		out LCD_DATA_DDR, r16

		; init EN pin
		in r16, LCD_EN_DDR
		ori r16, LCD_EN_MSK
		out LCD_EN_DDR, r16

		pop r16
		ret

	; Write LCD instruction byte prepared in r16.
	; Waits LCD_DELAY_US_INSTR us after instruction.
	; Used for all instructions except for ReturnHome and Clear.
	LCD_INSTR:
		push r16
		push r17
		push r18

		; clear RS bit
		in r18, LCD_RS_PORT
		andi r18, ~LCD_RS_MSK
		out LCD_RS_PORT, r18
		
		_LCD_INSTR_DATA:
		; prepare nibbles
		mov r17, r16
		andi r16, LCD_DATA_MSK	; low data nibble
		swap r17
		andi r17, LCD_DATA_MSK	; high data nibble

		; read data port
		in r18, LCD_DATA_PORT
		andi r18, ~LCD_DATA_MSK
		
		; load high data nibble		
		or r17, r18
		out LCD_DATA_PORT, r17
		rcall LCD_STROBE

		; load low data nibble		
		or r16, r18
		out LCD_DATA_PORT, r16
		rcall LCD_STROBE
		
		ldi r16, LCD_DELAY_US_INSTR
		rcall DELAYUS

		pop r18
		pop r17
		pop r16
		ret

	; Write LCD data byte prepared in r16 to DDRAM
	; Waits LCD_DELAY_US_INSTR ms after instruction.	
	LCD_DATA:
		; push registers on stack because of rjmp
		; which jumps in other routine code
		push r16
		push r17
		push r18
		
		; set RS bit
		in r18, LCD_RS_PORT
		ori r18, LCD_RS_MSK
		out LCD_RS_PORT, r18

		rjmp _LCD_INSTR_DATA

	; Write LCD instruction byte prepared in r16.
	; Waits LCD_DELAY_MS_CLR_RET ms after instruction.
	; Used for instructions ReturnHome and Clear.
	LCD_INSTR_DELAY:
		push r16		
		rcall LCD_INSTR
		ldi r16, LCD_DELAY_MS_CLR_RET
		rcall DELAYMS
		pop r16
		ret
	
	; LCD initialization to 4 bit mode
	LCD_INIT:
		push r16
		push r17
		push r18

		; initialize ports
		rcall LCD_PORT_INIT

		;-> Wait for more than 15 ms after VCC rises to 4.5 V
		ldi r16, LCD_DELAY_MS_INIT1
		rcall DELAYMS

		; clear RS bit
		in r16, LCD_RS_PORT
		andi r16, ~LCD_RS_MSK
		out LCD_RS_PORT, r16

		; read port unused values
		in r17, LCD_DATA_PORT
		andi r17, ~LCD_DATA_MSK
		
		; prepare instruction byte to r18
		ldi r18, LCD_INSTR_INIT_FNSET8B
		or r18, r17

		;-> Function set (Interface is 8 bits long.)
		out LCD_DATA_PORT, r18
		rcall LCD_STROBE

		;-> Wait for more than 4.1 ms
		ldi r16, LCD_DELAY_MS_INIT2
		rcall DELAYMS

		;-> Function set (Interface is 8 bits long.)
		out LCD_DATA_PORT, r18
		rcall LCD_STROBE

		;-> Wait for more than 100 us
		ldi r16, LCD_DELAY_US_INIT3
		rcall DELAYUS

		;-> Function set (Interface is 8 bits long.)
		out LCD_DATA_PORT, r18
		rcall LCD_STROBE
		; normal instruction wait
		ldi r16, LCD_DELAY_US_INSTR
		rcall DELAYUS

		;-> Function set (Set interface to be 4 bits long.)
		; Interface is 8 bits in length.

		; prepare instruction byte to r18
		ldi r18, LCD_INSTR_INIT_FNSET4B
		or r18, r17
		out LCD_DATA_PORT, r18
		rcall LCD_STROBE

		; normal instruction wait
		ldi r16, LCD_DELAY_US_INSTR
		rcall DELAYUS

		;-> Function set (Interface is 4 bits long.
		; Specify the number of display lines and character font.)
		; The number of display lines and character font
		; cannot be changed after this point.		
		ldi r16, LCD_INSTR_FNSET_4B_2L_58F
		rcall LCD_INSTR

		;-> Display off
		ldi r16, LCD_INSTR_DCTRL_OFF_NC_NB
		rcall LCD_INSTR

		;-> Display clear
		ldi r16, LCD_INSTR_CLRDISP
		rcall LCD_INSTR_DELAY

		;-> Entry mode set
		ldi r16, LCD_INSTR_ENTRY_INC_NS
		rcall LCD_INSTR

		;-> Display on
		ldi r16, LCD_INSTR_DCTRL_ON_NC_NB
		rcall LCD_INSTR

		pop r18
		pop r17
		pop r16
		ret

	; Helper macro that moves cursor on LCD to specified location
	; r16 register is used for LCD input 
	.macro lcd_goto	; row, col
		ldi r16, LCD_INSTR_SET_DDRAM | @0 + @1
		rcall LCD_INSTR
	.endmacro

	; Helper macro that writes specified byte on LCD
	; r16 register is used for LCD input 
	.macro lcd_write_bytei ; byteval
		ldi r16, @0
		rcall LCD_DATA
	.endmacro

	; Helper macro that writes specified register content on LCD
	; as byte value. r16 register is used for LCD input 
	.macro lcd_write_reg ; reg
		mov r16, @0
		rcall LCD_DATA
	.endmacro

	; Helper macro that writes 16 bit unsigned integer in r16:r17 on
	; lcd at specified position.
	; r16 register is used for LCD input 
	.macro lcd_write_int16 ; row, col
		rcall INT2STR16
		rcall REMLZEROS16
		lcd_goto @0, @1
		lcd_write_reg r4
		lcd_write_reg r3
		lcd_write_reg r2
		lcd_write_reg r1
		lcd_write_reg r0
	.endmacro		

	; Helper macro that writes 24 bit unsigned integer in r16:r17:r18 on
	; lcd at specified position.
	; r16 register is used for LCD input. 
	.macro lcd_write_int24 ; row, col
		rcall INT2STR24
		rcall REMLZEROS24
		lcd_goto @0, @1
		lcd_write_reg r7
		lcd_write_reg r6
		lcd_write_reg r5
		lcd_write_reg r4
		lcd_write_reg r3
		lcd_write_reg r2
		lcd_write_reg r1		
		lcd_write_reg r0
	.endmacro	

	; Helper macro that writes 24 bit unsigned integer in r16:r17:r18 on
	; lcd at specified position with one decimal digit.
	; r16 register is used for LCD input. 
	.macro lcd_write_int24_fp1 ; row, col
		rcall INT2STR24
		rcall REMLZEROS24_FP1
		lcd_goto @0, @1
		lcd_write_reg r6
		lcd_write_reg r5
		lcd_write_reg r4
		lcd_write_reg r3
		lcd_write_reg r2
		lcd_write_reg r1
		lcd_write_bytei '.'
		lcd_write_reg r0
	.endmacro	

	; Helper macro that writes 24 bit unsigned integer in r16:r17:r18 on
	; lcd at specified position with two decimal digits.
	; r16 register is used for LCD input.
	.macro lcd_write_int24_fp2 ; row, col
		rcall INT2STR24
		rcall REMLZEROS24_FP2
		lcd_goto @0, @1
		lcd_write_reg r6
		lcd_write_reg r5
		lcd_write_reg r4
		lcd_write_reg r3
		lcd_write_reg r2
		lcd_write_bytei '.'
		lcd_write_reg r1
		lcd_write_reg r0
	.endmacro

#endif /* _LCD_ASM_ */
