;=======================================================================;
;                                                                       ;
; Source code for RPM meter - simple device that measures revolutions   ;
; per minute, second and counts revolutions of rotating shaft with      ;
; encoder. Encoder scanning is based on CNY70 reflective optical sensor ;
; with transistor output. Black stripes of encoder generate pulses that ;
; rpm value is counted from. Device uses ATmega8 mcu, but this source   ;
; code can be modified and used with microcontrollers from Atmel AVR    ;
; family.                                                               ;
;                                                                       ;
; Target mcu:   Atmel AVR ATmega8                                       ;
; Mcu clock:    12 Mhz                                                  ;
; Author:       Viliam Sabol (sabolv@gmail.com)                         ;
;                                                                       ;
;=======================================================================;

.include "m8def.inc"

;=======================================================================;
; Interrupt vector definition                                           ;
; Reset Address is 0x0000 (BOOTRST fuse bit is unprogrammed)            ;
; Interrupt Vectors Start Address is 0x0001 (IVSEL in MCUCR is cleared) ;
;=======================================================================;

.org 0x0000				; Reset Handler
	; reti
	rjmp MAIN

.org 0x0001				; External Interrupt0 Handler
	reti
	; rjmp EXT_INT0

.org 0x0002				; External Interrupt1 Handler
	reti
	; rjmp EXT_INT1

.org 0x0003				; Timer2 Compare Handler
	reti
	; rjmp TIM2_COMP

.org 0x0004				; Timer2 Overflow Handler
	; reti
	rjmp TIM2_OVF

.org 0x0005				; Timer1 Capture Handler
	; reti
	rjmp TIM1_CAPT

.org 0x0006				; Timer1 CompareA Handler
	reti
	; rjmp TIM1_COMPA

.org 0x0007				; Timer1 CompareB Handler
	reti
	; rjmp TIM1_COMPB

.org 0x0008				; Timer1 Overflow Handler
	; reti
	rjmp TIM1_OVF

.org 0x0009				; Timer0 Overflow Handler
	reti
	; rjmp TIM0_OVF

.org 0x000A				; SPI Transfer Complete Handler
	reti
	; rjmp SPI_STC

.org 0x000B				; USART RX Complete Handler
	reti
	; rjmp USART_RXC

.org 0x000C				; UDR Empty Handler
	reti
	; rjmp USART_UDRE

.org 0x000D				; USART TX Complete Handler
	reti
	; rjmp USART_TXC

.org 0x000E				; ADC Conversion Complete Handler
	reti
	; rjmp ADC

.org 0x000F				; EEPROM Ready Handler
	reti
	; rjmp EE_RDY

.org 0x0010				; Analog Comparator Handler
	reti
	; rjmp ANA_COMP

.org 0x0011				; Two-wire Serial Interface Handler
	reti
	; rjmp TWSI

.org 0x0012				; Store Program Memory Ready
	reti
	; rjmp SPM_RDY

.org 0x0013

;=======================================================================;
; Global definitions                                                    ;
;=======================================================================;

	.equ MCUCLK 	= 12000000	; mcu frequency in Hz
	.equ LCDSLEEPMS = 100 		; time in milliseconds to sleep after displaying new measured value on LCD

	;======================================;
	; SRAM                                 ;
	;======================================;
	
	; Start data segment for variables in SRAM
	.dseg
	.org SRAM_START

		; Buttons
		BTN1_FLAGS:			.byte 1		; button 1 flag byte
		BTN1_COUNTER:		.byte 1		; button 1 counter byte
		BTN2_FLAGS:			.byte 1		; button 2 flag byte		
		BTN2_COUNTER:		.byte 1		; button 2 counter byte
		BTN2_REPCNT:		.byte 1		; repeat counter used for triggering button when it is held down

		; Device mode
		DEVMODE:			.byte 1		; current device mode byte

		; RPM/RPS meter
		FPMODE:				.byte 1		; current floating point mode		
		TMCONST_FLAG:		.byte 1		; flag for recounting time constants
		ENC_TICKS_SRAM:		.byte 1		; encoder ticks value read from eeprom

		INT_TICKS_FP0_SEC:	.byte 3		; number of ticks between ICP interrupts for ENCODER_TICKS per min(sec)
		INT_TICKS_FP1_SEC:	.byte 3		; number of ticks between ICP interrupts for ENCODER_TICKS per min(sec) * 10
		INT_TICKS_FP2_SEC:	.byte 3		; number of ticks between ICP interrupts for ENCODER_TICKS per min(sec) * 100

		INT_TICKS_FP0_MIN:	.byte 3		; number of ticks between ICP interrupts for ENCODER_TICKS per min(sec)
		INT_TICKS_FP1_MIN:	.byte 3		; number of ticks between ICP interrupts for ENCODER_TICKS per min(sec) * 10
		INT_TICKS_FP2_MIN:	.byte 3		; number of ticks between ICP interrupts for ENCODER_TICKS per min(sec) * 100
		
		FPMODE_MIN_CNT:		.byte 1		; number of floating point modes for rpm mode

	;======================================;
	; EEPROM                               ;
	;======================================;
	
	; Start data segment for variables in EEPROM
	.eseg

		ENC_TICKS:	.db DEF_ENC_TICKS 	; encoder ticks value in eeprom

	; Code segment
	.cseg

	;======================================;
	; Device mode                          ;
	;======================================;

	.equ RPM_MODE 	= 0
	.equ RPS_MODE 	= 1
	.equ CNT_MODE 	= 2
	.equ ENC_MODE	= 3
	.equ MODE_COUNT = 3	; number of device modes

	.equ FPMODE_STEP	= 3
	.equ FPMODE_0 		= 0 * FPMODE_STEP ; without floating point
	.equ FPMODE_1 		= 1 * FPMODE_STEP ; one decimal digit
	.equ FPMODE_2 		= 2 * FPMODE_STEP ; two decimal digits
	.equ FPMODE_COUNT 	= 3 * FPMODE_STEP ; number of floating point modes

	.equ TMCONST_RECOUNT= 255 ; value of TIMECONST_FLAG byte in sram that signalizes that values for rpm/rps need to be updated

	;======================================;
	; Buttons                              ;
	;======================================;
	
	.equ BTN1_PIN 	= PD7   ; button 1 pin definition
	.equ BTN1_DDR 	= DDRD  ; button 1 pin data direction register
	.equ BTN1_PORT 	= PORTD ; button 1 pin port register
	.equ BTN1_INPUT	= PIND  ; button 1 input pin register

	.equ BTN2_PIN 	= PD6   ; button 2 pin definition
	.equ BTN2_DDR 	= DDRD  ; button 2 pin data direction register
	.equ BTN2_PORT 	= PORTD ; button 2 pin port register
	.equ BTN2_INPUT	= PIND  ; button 2 input pin register

	;======================================;
	; Timer1 input capture / RPM Counter   ;
	;======================================;

	.equ ICP_PIN 	= PB0	; input capture pin	definition
	.equ ICP_DDR 	= DDRB  ; input capture pin data direction register
	.equ ICP_PORT 	= PORTB ; input capture pin port register

	.def TCNT13 = r9	; TCNT13 register is 3rd byte for timer1
						; that is used with this register as 24 bit counter

	.def TSTAMP0 = r10	; 24bit timestamp for Input Capture of timer1
	.def TSTAMP1 = r11	; used as time between two rising edges on
	.def TSTAMP2 = r12	; input capture pin ICP_PIN
	
	.equ TIMER1_PRESCALER = 256 ; timer1 prescaler value
	
	.equ SECOND = MCUCLK / TIMER1_PRESCALER ; ticks of timer1 with prescaler clk/256 in one second
	.equ MINUTE = SECOND * 60               ; ticks of timer1 with prescaler clk/256 in one minute
	.equ INT_TICKS_FP2_MAX = 167772					; maximum value for INT_TICKS_FP0, to be able to count INT_TICKS_FP2
	.equ INT_TICKS_FP1_MAX = INT_TICKS_FP2_MAX * 10	; maximum value for INT_TICKS_FP0, to be able to count INT_TICKS_FP1

	.equ TCNT13_TIMEOUT = 3 ; number of overflow interrupts of 16bit timer1
							; with prescaler clk/256 to clear timestamp values and
							; timer1 counter. When timeout expires, zero value is
							; displayed as rpm/rps result. Timeout is valid only when
							; device is in rpm/rps mode.
							; Timeout is 3 * 256(prescaler) * 65536(16bit counter) / 12 000 000(clk) = 4.1s

	.equ DEF_ENC_TICKS =  1 ; default number of encoder ticks, one revolution = ENC_TICKS number of input interrupts
	.equ MAX_ENC_TICKS = 64
	.equ MIN_ENC_TICKS =  1	

	; Helper macro used to reset rmp-timestamp register values.
	.macro clrstamprpm
		clr TSTAMP0
		dec TSTAMP0
		clr TSTAMP1
		dec TSTAMP1
		clr TSTAMP2
		dec TSTAMP2
	.endmacro

	; Helper macro used to reset rev.-counter register values.
	.macro clrstampcnt
		clr TSTAMP0
		clr TSTAMP1
		clr TSTAMP2
	.endmacro

;=======================================================================;
; Library includes                                                      ;
;=======================================================================;

.include "delay.asm"
.include "auxiliary.asm"
.include "lcd.asm"
.include "arithmetic.asm"

;=======================================================================;
; Interrupt handlers                                                    ;
;=======================================================================;

	; 8bit Timer2 overflow handler is used for buttons debouncing.
	TIM2_OVF:
		
		; save status register and tmp registers used by handler
		push r16
		in r16, SREG
		push r16
		push r17
		push r18
		push r19
		push r20
		
		ldi r19, BTN_RISING_EDGE_MSK ; load rising edge mask for button 1
		clr r20						 ; clear repeat counter for button 1

		; handle button 1
		in r17, BTN1_INPUT			; read button value to r17(BTN_NEWSTATE) register for debounce function
		com r17						; invert bit values because pin value is low when button is pressed
		andi r17, _BV(BTN1_PIN)		
		ldsram r16, BTN1_FLAGS		; prepare flags to r16(BTN_FLAGS) register for debounce function
		ldsram r18, BTN1_COUNTER	; prepare counter to r18(BTN_COUNTER) register for debounce function
		rcall BUTTON_DEBOUNCE
		stsram BTN1_FLAGS, r16		; store values of button back to sram
		stsram BTN1_COUNTER, r18

		ldsram r16, DEVMODE			; load rising edge mask for second button with repeat only in ENC_MODE
		cpi r16, ENC_MODE
		brne _TIM2_OVF_BTN2_MSK
		ldi r19, BTN_RISREP_EDGE_MSK
		_TIM2_OVF_BTN2_MSK:

		; handle button 2		
		in r17, BTN2_INPUT			; read button value to r17(BTN_NEWSTATE) register for debounce function
		com r17						; invert bit values because pin value is low when button is pressed
		andi r17, _BV(BTN2_PIN)		
		ldsram r16, BTN2_FLAGS		; prepare flags to r16(BTN_FLAGS) register for debounce function
		ldsram r18, BTN2_COUNTER	; prepare counter to r18(BTN_COUNTER) register for debounce function
		ldsram r20, BTN2_REPCNT		; prepare repeat counter to r20(BTN_REPCNT) register for debounce function
		rcall BUTTON_DEBOUNCE
		stsram BTN2_FLAGS, r16		; store values of button back to sram
		stsram BTN2_COUNTER, r18
		stsram BTN2_REPCNT, r20

		ldsram r17, BTN1_FLAGS		; if button 2 is pressed while button i is held, change mode to encoder ticks selection
		mov r18, r16
		lsl r18
		brcc _TIM2_OVF_BTN1
		mov r18, r17
		and r18, r16
		lsl r18
		lsl r18
		brcc _TIM2_OVF_BTN1		
		rcall ENC_MODE_SET
		rjmp _TIM2_OVF_RET

		_TIM2_OVF_BTN1:
		lsl r17						; shift edge bit to carry
		brcc _TIM2_OVF_BTN2			; handle next button if edge not detected
		rcall DEV_MODE_CHANGE		; change device mode on button 1 press

		_TIM2_OVF_BTN2:
		lsl r16						; shift edge bit to carry
		brcc _TIM2_OVF_RET			; return from interrupt if edge not detected
		rcall DEV_MODE_FN			; call alternate mode function on button 2 press

		_TIM2_OVF_RET: 				; return from interrupt
		pop r20						; restore status register and tmp registers
		pop r19
		pop r18
		pop r17
		pop r16
		out SREG, r16
		pop r16
		reti	
		
	; 16bit Timer1 overflow handler used to
	; increment TCNT13 register - third byte of 24 bit counter with TCNT1L and TCNT1H
	TIM1_OVF:
		push r16
		in r16, SREG
		push r16

		inc TCNT13		
		
		; is current device mode counter or higher?
		ldsram r16, DEVMODE
		cpi r16, CNT_MODE
		brsh _TIM1_OVF_RET

		; if it is rpm/rps mode test timeout value
		mov r16, TCNT13
		cpi r16, TCNT13_TIMEOUT
		brne _TIM1_OVF_RET

		; if timeout expired clear timestamp ...
		clrstamprpm
		; and clear Timer1 counter
		clr TCNT13
		out TCNT1H, TCNT13
		out TCNT1L, TCNT13

		_TIM1_OVF_RET:
		pop r16
		out SREG, r16
		pop r16
		reti

	; Input capture interrupt handler for Timer1.
	; If device is in RMP_MODE/RPS_MODE, actual number of counter ticks between two interrupts are stored to
	; TSTAMP0:2 registers.
	; If device is in CNT_MODE, TSTAMP0:1 registers are incremented on every input capture interrupt.
	TIM1_CAPT:
		push r16
		in r16, SREG
		push r16

		; is current device mode counter ?
		ldsram r16, DEVMODE
		cpi r16, CNT_MODE
		brne _TIM1_CAPT_STAMP
		
		; counter mode - increment TSTAMP0:2
		ldi r16, 1
		add TSTAMP0, r16
		clr r16
		adc TSTAMP1, r16
		adc TSTAMP2, r16
		rjmp _TIM1_CAPT_RET

		; rpm/rps mode - capture timestamp to TSTAMP0:2
		_TIM1_CAPT_STAMP:
		; capture timestamp
		mov TSTAMP2, TCNT13
		in TSTAMP0, ICR1L
		in TSTAMP1, ICR1H

		; clear Timer1 counter
		clr TCNT13
		out TCNT1H, TCNT13
		out TCNT1L, TCNT13
		
		_TIM1_CAPT_RET:
		pop r16
		out SREG, r16
		pop r16
		reti

;=======================================================================;
; Peripheral initialization functions                                   ;
;=======================================================================;

	; Initialize Timer2 to overflow interrupt with
	; prescaler clk/256, so it will trigger an interrupt
	; every 5.4 ms for clk = 12MHz. Timer is used for 
	; button debouncing.
	TIMER2_INIT:
		push r16
		clr r16
		out TCCR2, r16	; stop timer2
		out TCNT2, r16	; clear counter byte

		; clear pending interrupts for overflow
		in r16, TIFR
		ldi r16, _BV(TOV2)
		out TIFR, r16

		; set interrupt mask for overflow interrupt
		in r16, TIMSK
		ori r16, _BV(TOIE2)
		out TIMSK, r16

		; select clk/256 prescaler
		ldi r16, _BV(CS22) | _BV(CS21)
		out TCCR2, r16
		pop r16
		ret

	; Initialize 16bit Timer1 to input capture mode
	; and enable timer overflow interrupt to utilize it as 24 bit counter.
	TIMER1_INIT:
		push r16
		clr r16
		out TCCR1B, r16	; stop timer1
		out TCNT1H, r16	; clear counter second byte
		out TCNT1L, r16 ; first byte
		clr TCNT13		; third byte

		; clear pending interrupts for input capture and overflow
		in r16, TIFR
		ori r16, _BV(ICF1) | _BV(TOV1)
		out TIFR, r16

		; set interrupt masks for input capture and overflow interrupts
		in r16, TIMSK
		ori r16, _BV(TICIE1) | _BV(TOIE1)
		out TIMSK, r16

		; select clk/256 prescaler, and
		; input capture rising edge sense
		ldi r16, _BV(CS12) | _BV(ICES1) ;| _BV(ICNC1)
		out TCCR1B, r16
		pop r16
		ret	

;=======================================================================;
; Buttons and debouncing                                                ;
;=======================================================================;

	; Initialization of buttons and debouncing fuction. 
	BTN_INIT:
		push r16
		clr r16

		; Button 1
		cbi BTN1_DDR, BTN1_PIN		; initialize button pins and data direction
		sbi BTN1_PORT, BTN1_PIN		; enable pull-ups
		stsram BTN1_FLAGS, r16		; clear button flags variable
		stsram BTN1_COUNTER, r16	; clear button debouncing counter
		
		; Button 2
		cbi BTN2_DDR, BTN2_PIN		; initialize button pins and data direction
		sbi BTN2_PORT, BTN2_PIN		; enable pull-ups
		stsram BTN2_FLAGS, r16		; clear button flags variable
		stsram BTN2_COUNTER, r16	; clear button debouncing counter
		stsram BTN2_REPCNT, r16		; clear button repeat counter
		
		rcall TIMER2_INIT			; start debouncing timer2
		
		pop r16
		ret

	; Debouncing routine for buttons. If button real state(call it temporary state)
	; does not change for BTN_BOUNCECNT measures, button is in stable state. To count
	; nuber of measures in which button does not change state button counter is used.
	; When rising edge on button input is detected, msb of stable button state is set,
	; otherwise that bit is cleared.

	.equ BTN_TMPSTAT_MSK = 0b00000001	; mask for temporary state in button flags register
	.equ BTN_STBSTAT_MSK = 0b01000000	; mask for stable state in button flags register
	.equ BTN_EDGE_MSK   = 0b10000000	; mask for edge in input button signal
	.equ BTN_RISING_EDGE_MSK  = 0b00000001	; BTN_FLAGS register value on rising edge (when state is stable) - new state is 1, old stable state is 0
	.equ BTN_RISREP_EDGE_MSK  = 0b01000001	; BTN_FLAGS register value on rising edge with repating (when state is stable) - new state is 1, old stable state is 1, but it reacts also on real rising edge
	.equ BTN_FALLING_EDGE_MSK = 0b01000000	; BTN_FLAGS register value on falling edge (when state is stable) - new state is 0, old stable state is 1	
	.equ BTN_BOUNCECNT   = 8 			; number of measures that button cannot change state to consider it is in stable state
	.equ BTN_REPEATCNT	 = 8
	.def BTN_FLAGS	 = r16
	.def BTN_NEWSTAT = r17
	.def BTN_COUNTER = r18
	.def BTN_EDGE_SN = r19
	.def BTN_REPCNT = r20

	; Input: BTN_FLAGS register contains last temporary button state in bit0
	;				   and last stable buton state in bit6. Bit7 is rising edge detector.
	;        BTN_NEWSTAT register contains actual state of button read from button port, register is nonzero when button is pressed, otherwise zero.
	;        BTN_COUNTER register contains state counter for button.
	;		 BTN_EDGE_SN register contains value that is compared with BTN_FLAGS at every stable state.
	;					 if values are equal, BTN_EDGE_MSK bit in BTN_FLAGS register is set. To detect rising edge,
	;					 use BTN_RISING_EDGE_MSK value for BTN_EDGE_SN. For falling edge use BTN_FALLING_EDGE_MSK.
	; 		 BTN_REPCNT repeat counter for button
	; 
	; Output: BTN_FLAGS register contains new button states. Moreover when rising edge
	;                   on button input is detected, msb of stable button state is set,
	;                   otherwise that bit is cleared.
	;         BTN_COUNTER new value of button counter.
	;         BTN_NEWSTAT value in register is not preserved
	BUTTON_DEBOUNCE:
		push r21

		mov r21, BTN_FLAGS				; load button last temporary state to r19
		andi r21, BTN_TMPSTAT_MSK

		tst BTN_NEWSTAT
		breq _BUTTON_DEBOUNCE_CMP
		ldi BTN_NEWSTAT, 1				; if new state was nonzero load 1 to register

		_BUTTON_DEBOUNCE_CMP:
		andi BTN_FLAGS, ~(BTN_TMPSTAT_MSK | BTN_EDGE_MSK) ; clear edge bit and last temp state
		or BTN_FLAGS, BTN_NEWSTAT		; write new state to button flags
		cp BTN_NEWSTAT, r21				; compare new state and last temporary state
		breq _BUTTON_DEBOUNCE_NOCHAN			
										; state of button was changed
		clr BTN_COUNTER					; clear button counter
		rjmp _BUTTON_DEBOUNCE_RET		; return

		_BUTTON_DEBOUNCE_NOCHAN:		; state of button was not changed
		inc BTN_COUNTER					; increment button counter
		cpi BTN_COUNTER, BTN_BOUNCECNT	; compare if state is not changed after BTN_BOUNCECNT measures
		brne _BUTTON_DEBOUNCE_RET
										; new state is stable
		clr BTN_COUNTER					; clear button counter
		cp BTN_FLAGS, BTN_EDGE_SN		; check if edge in input signal occured
		breq _BUTTON_DEBOUNCE_REPEAT
				
		mov r21, BTN_EDGE_SN			; if in repeat mode, generate edge on
		andi r21, ~BTN_STBSTAT_MSK		; button signal rising edge too
		cp BTN_FLAGS, r21
		brne _BUTTON_DEBOUNCE_ACTUALIZE
		rjmp _BUTTON_DEBOUNCE_EDGE

		_BUTTON_DEBOUNCE_REPEAT:			; if in repeat mode, increment counter
		cpi BTN_EDGE_SN, BTN_RISREP_EDGE_MSK; and if BTN_REPEATCNT value is in counter
		brne _BUTTON_DEBOUNCE_EDGE			; generate edge on button
		inc BTN_REPCNT
		cpi BTN_REPCNT, BTN_REPEATCNT
		brne _BUTTON_DEBOUNCE_ACTUALIZE
		clr BTN_REPCNT

		_BUTTON_DEBOUNCE_EDGE:
		ori BTN_FLAGS, BTN_EDGE_MSK		; edge occured, set BTN_EDGE_MSK flag

		_BUTTON_DEBOUNCE_ACTUALIZE:		; actualize new stable state
		andi BTN_FLAGS, ~BTN_STBSTAT_MSK; clear stable state
		tst BTN_NEWSTAT
		breq _BUTTON_DEBOUNCE_RET
		ori BTN_FLAGS, BTN_STBSTAT_MSK	; set stable state if BTN_NEWSTAT == 1

		_BUTTON_DEBOUNCE_RET:			; return from routine
		pop r21
		ret

;=======================================================================;
; Helper functions                                                      ;
;=======================================================================;

	; Change mode to ENC_MODE, or exit from it	
	; EEPROM write is called so interrupts should be disabled
	ENC_MODE_SET:
		push r16
		push r17
		push r18
		
		ldsram r16, DEVMODE		
		cpi r16, ENC_MODE
		breq _ENC_MODE_SET_EXIT
		; entering to ENC_MODE		
		ldi r16, ENC_MODE
		stsram DEVMODE, r16
		rjmp _ENC_MODE_SET_RET

		_ENC_MODE_SET_EXIT:
		; exitting from ENC_MODE
		ldsram r16, ENC_TICKS_SRAM
		eepromwrite ENC_TICKS	; store ENC_TICKS from sram to EEPROM
		ldi r16, RPM_MODE
		stsram DEVMODE, r16		; reset device mode
		clrstamprpm				; clear timestamps
		ldi r16, FPMODE_0
		stsram FPMODE, r16		; reset floating point mode
		stsrami TMCONST_FLAG, TMCONST_RECOUNT	; signalize that values for rpm/rps 
												; are invalid and need to be recounted

		_ENC_MODE_SET_RET:
		pop r18
		pop r17
		pop r16
		ret

	; Change device mode. If current mode is last, set first mode.
	DEV_MODE_CHANGE:
		push r18
				
		ldsram r18, DEVMODE
		cpi r18, ENC_MODE
		breq _DEV_MODE_CHANGE_RET
		stsrami FPMODE, FPMODE_0	; reset FMODE to FPMODE_0
		inc r18						; cycle through device states, devmode is in r18
		cpi r18, MODE_COUNT
		brne _DEV_MODE_CHANGE_STORE	; if it was last state, set 0(first state) as next state
		clr r18

		_DEV_MODE_CHANGE_STORE:
		stsram DEVMODE, r18
		
		clrstamprpm					; clear timestamp registers
		cpi r18, CNT_MODE			; if current mode is device counter, clear timestamp
									; that is initialized for rpm mode where first result value should be zero
		brne _DEV_MODE_CHANGE_RET
		clrstampcnt
		
		_DEV_MODE_CHANGE_RET:
		pop r18
		ret

	; Device alternate function button handler
	DEV_MODE_FN:
		push r18
		push r19

		ldsram r18, DEVMODE		; in rpm mode change FPMODE 
		cpi r18, RPM_MODE		; with control of FPMODE_MIN_CNT
		brne _DEV_MODE_FN_RPS
		ldi r18, FPMODE_STEP
		ldsram r19, FPMODE
		add r19, r18
		ldsram r18, FPMODE_MIN_CNT
		cp r19, r18
		brne _DEV_MODE_FN_STORE
		ldi r19, FPMODE_0
		rjmp _DEV_MODE_FN_STORE

		_DEV_MODE_FN_RPS:		; in rps mode change FPMODE
		cpi r18, RPS_MODE		; without control of FPMODE_MIN_CNT
		brne _DEV_MODE_FN_CNT
		ldi r18, FPMODE_STEP
		ldsram r19, FPMODE
		add r19, r18		
		cpi r19, FPMODE_COUNT
		brne _DEV_MODE_FN_STORE
		ldi r19, FPMODE_0
		rjmp _DEV_MODE_FN_STORE

		_DEV_MODE_FN_CNT:
		cpi r18, CNT_MODE
		brne _DEV_MODE_FN_ENC
		clrstampcnt				; in cnt mode only clear timestamp
		rjmp _DEV_MODE_FN_RET
		
		_DEV_MODE_FN_ENC:
		cpi r18, ENC_MODE
		brne _DEV_MODE_FN_RET
		; increment encoder ticks variable and reset it if it is maximum
		ldsram r18, ENC_TICKS_SRAM
		cpi r18, MAX_ENC_TICKS
		breq _DEV_MODE_FN_ENC_RST
		inc r18
		stsram ENC_TICKS_SRAM, r18
		rjmp _DEV_MODE_FN_RET
		_DEV_MODE_FN_ENC_RST:
		ldi r18, MIN_ENC_TICKS
		stsram ENC_TICKS_SRAM, r18
		rjmp _DEV_MODE_FN_RET

		_DEV_MODE_FN_STORE:
		stsram FPMODE, r19		; store new FPMODE in r19 to sram
		clrstamprpm				; clear timestamp for rpm/rps mode

		_DEV_MODE_FN_RET:
		pop r19
		pop r18
		ret

	; Initialization routine for device
	; Stack pointer must be initialized
	INIT:
		cbi ICP_DDR, ICP_PIN		; initialize ICP pin as input
		rcall TIMER1_INIT			; initialize input capture timer1
		clrstamprpm					; clear timestamp registers

		rcall LCD_INIT				; initialize LCD

		rcall BTN_INIT				; initialize buttons

		eepromread ENC_TICKS		; read ENC_TICKS value from eeprom and store it to sram,
		cpi r16, MAX_ENC_TICKS + 1	; check if value is not higher then maximum
		brlo _INIT_ENC_TICKS_OK
		ldi r16, MAX_ENC_TICKS
		_INIT_ENC_TICKS_OK:
		stsram ENC_TICKS_SRAM, r16	; interrupts are disabled during initialization
		
		stsrami DEVMODE, RPM_MODE	; initialize device mode
		stsrami FPMODE, FPMODE_0	; initialize floating point mode
		stsrami TMCONST_FLAG, TMCONST_RECOUNT	; signalize that values for rpm/rps 
												; are invalid and need to be recounted
		ret
	
	; Helper macro used to initialize time constants before measuring.
	; value without fp is stored in r16:r18 and is not modified
	.macro recount_sec_fp_constants ; MAX_INPUT_VALUE, endlabel, multiplicator, ramstore
		; count constant for one decimal digit measures
		ldi r20,  high(@0)
		ldi r21, byte3(@0)
		cpi r16,   low(@0)
		cpc r17, r20
		cpc r18, r21
		brsh @1			
		; input value < MAX_INPUT_VALUE, so we can multiply it without overflow
		ldi r19, @2
		rcall MUL24U8
		stsram24u @3, r19, r20, r21	; store constant to sram
		@1:
	.endmacro

	; Helper macro used to initialize time constants before measuring.
	; value without fp is stored in r16:r18 and is not modified
	.macro recount_min_fp_constants ; MAX_INPUT_VALUE, countlabel, retlabel, multiplicator, ramstore, fpmode_counter, fpmode_max
		ldi r20,  high(@0)
		ldi r21, byte3(@0)
		cpi r16,   low(@0)
		cpc r17, r20
		cpc r18, r21
		brlo @1
		; input value >= MAX_INPUT_VALUE so disable floating point mode
		stsrami @5, @6
		rjmp @2
		@1:
		; input value < MAX_INPUT_VALUE, so we can multiply it without overflow
		ldi r19, @3
		rcall MUL24U8
		stsram24u @4, r19, r20, r21	; store constant for one decimal measures to sram
	.endmacro

	; Counts time constants for all floating point modes. These values are stored in sram and device
	; needs not to count them for every measure cycle.
	RECOUNT_TMCONSTANTS:		
		clr r16
		stsram TMCONST_FLAG, r16		; clear recount flag
		
		; count rps constants
		ldi24u r16, r17, r18, SECOND	; prepare SECOND as dividend
		ldsram r19, ENC_TICKS_SRAM		; prepare encoder ticks as divisor
		clr r20
		clr r21
		rcall DIV24U					; count constant for rps measure
		stsram24u INT_TICKS_FP0_SEC, r16, r17, r18	; store constant to sram
		
		; count constant for measure with one decimal digit
		recount_sec_fp_constants INT_TICKS_FP1_MAX, _RECOUNT_SEC_FP1_COUNT_END, 10, INT_TICKS_FP1_SEC 
		; count constant for measure with two decimal digits
		recount_sec_fp_constants INT_TICKS_FP2_MAX, _RECOUNT_SEC_FP2_COUNT_END, 100, INT_TICKS_FP2_SEC 

		; count rpm constants
		stsrami FPMODE_MIN_CNT, FPMODE_2 + FPMODE_STEP	; init number of floating point modes for rpm
		ldi24u r16, r17, r18, MINUTE	; prepare MINUTE as dividend
		ldsram r19, ENC_TICKS_SRAM		; prepare encoder ticks as divisor
		clr r20
		clr r21
		rcall DIV24U					; count constant for rpm measure
		stsram24u INT_TICKS_FP0_MIN, r16, r17, r18	; store constant to sram
		
		; count constant for measure with one decimal digit
		recount_min_fp_constants INT_TICKS_FP1_MAX, _RECOUNT_MIN_FP1_COUNT, _RECOUNT_RET, 10, INT_TICKS_FP1_MIN, FPMODE_MIN_CNT, FPMODE_1 
		; count constant for measure with two decimal digits
		recount_min_fp_constants INT_TICKS_FP2_MAX, _RECOUNT_MIN_FP2_COUNT, _RECOUNT_RET, 100, INT_TICKS_FP2_MIN, FPMODE_MIN_CNT, FPMODE_2
		_RECOUNT_RET:
		ret
		
;=======================================================================;
; Main program                                                          ;
;=======================================================================;
	
	.def fpmodecpy = r8	; actual copy of floating point mode register

	MAIN:

		; Stack init at top of RAM
		ldi r16, high(RAMEND)
		out SPH,r16
		ldi r16, low(RAMEND)
		out SPL, r16		

		; Initialization
		rcall INIT

		; Enable global interrupts
		sei

		LOOP:

			ldi r16, LCDSLEEPMS				; LCD delay
			rcall DELAYMS

			ldsram r16, TMCONST_FLAG
			cpi r16, TMCONST_RECOUNT		; get TMCONST_FLAG and decide if recount of values is needed
			brne _LOOP_DEV_MODE				; if not, proceed to state selection
			rcall RECOUNT_TMCONSTANTS		; recount time constants
					

			_LOOP_DEV_MODE:					; select device mode
			lcd_goto LCD_DDRAM_LINE0, 0		; set starting position on LCD in left upper corner
			ldsram fpmodecpy, FPMODE		; copy FPMODE
			ldsram r16, DEVMODE				; copy DEVMODE
			cpi r16, RPM_MODE
			brne _LOOP_RPS_MODE
											; rpm mode
			lcd_write_bytei 'r'
			lcd_write_bytei 'p'
			lcd_write_bytei 'm'
			ldsram24uor r16, r17, r18, INT_TICKS_FP0_MIN, fpmodecpy	; load time constant to r16:r18
			rjmp _LOOP_REV_CNT

			_LOOP_RPS_MODE:
			cpi r16, RPS_MODE
			brne _LOOP_COUNTER_MODE
											; rps mode
			lcd_write_bytei 'r'
			lcd_write_bytei 'p'
			lcd_write_bytei 's'			
			ldsram24uor r16, r17, r18, INT_TICKS_FP0_SEC, fpmodecpy	; load time constant to r16:r18
			rjmp _LOOP_REV_CNT

			_LOOP_COUNTER_MODE:
			cpi r16, CNT_MODE
			brne _LOOP_ENC_MODE
											; counter mode
			lcd_write_bytei 'c'
			lcd_write_bytei 'n'
			lcd_write_bytei 't'
			
			cli								; Load counter ticks
			mov r16, TSTAMP0
			mov r17, TSTAMP1
			mov r18, TSTAMP2
			sei
			ldi r19, 100
			rcall MUL24U8
			mov r16, r19
			mov r17, r20
			mov r18, r21

			ldsram r19, ENC_TICKS_SRAM
			clr r20
			clr r21
			rcall DIV24U
			lcd_write_int24_fp2 LCD_DDRAM_LINE1, 0	; write result with two decimal digits
			rjmp LOOP			
			
			_LOOP_ENC_MODE:
											; encoder ticks selection mode
			lcd_write_bytei 'e'
			lcd_write_bytei 'n'
			lcd_write_bytei 'c'
			ldsram r16, ENC_TICKS_SRAM
			clr r17
			clr r18
			lcd_write_int24 LCD_DDRAM_LINE1, 0
			rjmp LOOP
			
			_LOOP_REV_CNT:					
			cli								; Load actual time between revolutions
			mov r19, TSTAMP0
			mov r20, TSTAMP1
			mov r21, TSTAMP2
			sei			
			rcall DIV24U					; Count RPM to r16:r17:r18
			ldi r19, FPMODE_0
			cp fpmodecpy, r19
			brne _LOOP_FPMODE_1
			lcd_write_int24 LCD_DDRAM_LINE1, 0	; write result without floating point
			rjmp LOOP
			_LOOP_FPMODE_1:
			ldi r19, FPMODE_1
			cp fpmodecpy, r19
			brne _LOOP_FPMODE_2
			lcd_write_int24_fp1 LCD_DDRAM_LINE1, 0	; write result with one decimal digit
			rjmp LOOP
			_LOOP_FPMODE_2:
			lcd_write_int24_fp2 LCD_DDRAM_LINE1, 0	; write result with two decimal digits
			rjmp LOOP
